use strict;

sub read_pnpdata {
# get the input file name
  my($pnpdata) = @_;
  my($pnp);
# if no input file specified, try to get data from pnpdump itself
# must be root to run pnpdump - we'll see if we have execute access, anyway
  my($datasrc);
  if($pnpdata eq "") {
    $datasrc = "/sbin/pnpdump|";
    if(! -x "/sbin/pnpdump") {
      return 0;
    }
  }
  else {
    $datasrc = "<$pnpdata";
  }
# open our data source of choice
  if(!open(PNPDATA, $datasrc)) {
    return 0;
  }

# assume boards found unless we determine otherwise
  $pnp->{boardsfound} = 1;
  $pnp->{agent_ver} = " unknown";
  $pnp->{agent_flags} = "NONE";
  $pnp->{verifyld} = 1;

  $_ = <PNPDATA>;
  if(/^# \$Id: \S* (\S*) .*/) { $pnp->{agent_ver} = $1; }
  else { close(PNPDATA); return -2; }
  while(!eof(PNPDATA)) {
    $_ = <PNPDATA>;
# get the version of the pnpdump read agent that generated this output
    if(/^# Compiler flags: (.*)/) { $pnp->{agent_flags} = $1; }
# find out the port address we're reading from
    elsif(/^# Trying port address (.*)/) { $pnp->{readport} = hex($1); }
# if we haven't found any boards, we'll know it now
    elsif(/^# No boards found/) { $pnp->{boardsfound} = 0; }
# the beginning of a card info entry
    elsif(/^# Card (.*): .*/) {
# reference our data structure, to simplify calling it
      my($board) = \%{$pnp->{board}[$1]};
      $board->{first_ld} = -1;
      until((($_ = <PNPDATA>) =~ /^# End tag.*/) || eof(PNPDATA)) {
        if(/^# Vendor Id (\S*), Serial Number (\S*), /) {
          $board->{vendorid} = $1;
          $board->{serialnum} = $2;
        }
        elsif(/^# Vendor Id (\S*), No Serial Number \(-1\), /) {
          $board->{vendorid} = $1;
          $board->{serialnum} = -1;
        }
# this is for the old style (v1.10 for sure) pnpdump output
        elsif(/^# (\S*) Serial No (\S*) /) {
          $board->{vendorid} = $1;
          $board->{serialnum} = $2;
        }
# the card name as it appears in the ROM on the card
        elsif(/^#\s*ANSI string -->(.*)<--/) {
          $board->{name} = $1;
        }
# the start of our logical device definition
        elsif(/^\(CONFIGURE $board->{vendorid}\/$board->{serialnum} \(LD (\S*)/) {
          if($board->{first_ld} == -1) {
            $board->{first_ld} = $1;
          }
          my($ld) = \%{$board->{ld}[$1]};
          my($prof) = 0;
          $ld->{profile}[$prof]{priority} = "default";
          my($profile) = \%{$ld->{profile}[$prof]};
          my($io) = 0;
          my($mem) = 0;
          my($irq) = 0;
          my($dma) = 0;
          until((($_ = <PNPDATA>) =~ /^# \(ACT Y\)/) || eof(PNPDATA))
          {
# the start of a profile/schema for device setup
            if(/^#\s*Start dependent functions: priority (\S*)/) {
              $profile = \%{$ld->{profile}[$prof]};
              $profile->{priority} = $1;
              $prof++;
              $io = 0;
              $mem = 0;
              $irq = 0;
              $dma = 0;
            }
# the name string for the logical device
            elsif(/^#\s*ANSI string -->(.*)<--/) {
              $ld->{name} = $1;
            }
# an IRQ channel definition
            elsif(/^#\s*IRQ (.*)\./) {
              my($irqs);
              ($irqs = $1) =~ s/ or/,/;
              @{$profile->{irq}[$irq]{irqs}} = split(', ', $irqs);
              until((($_ = <PNPDATA>) =~ /^# \(INT $irq .*/) || eof(PNPDATA)) {
                if(/^#\s*(\S*) true, (\S*) sensitive interrupt/) {
                  my($type);
                  ($type = "$1 $2") =~ tr/A-Z/a-z/;
                  if($type =~ "high edge") {
                    $profile->{irq}[$irq]{mode} = "+E";
                  }
                  elsif($type =~ "low edge") {
                    $profile->{irq}[$irq]{mode} = "-E";
                  }
                  elsif($type =~ "high level") {
                    $profile->{irq}[$irq]{mode} = "+L";
                  }
                  elsif($type =~ "low level") {
                    $profile->{irq}[$irq]{mode} = "-L";
                  }
                }
              }
              $irq++;
            }
# a DMA channel definition
            elsif(/^#.*DMA channel (.*)\./) {
              my($dmas);
              ($dmas = $1) =~ s/ or/,/;
              @{$profile->{dma}[$dma]{dmas}} = split(', ', $dmas);
              until((($_ = <PNPDATA>) =~ /^# \(DMA $dma .*/) || eof(PNPDATA)) {
              }
              $dma++;
            }
# an IO window definition
            elsif(/^#\s*Logical device decodes (\S*) bit IO address lines/) {
              $profile->{io}[$io]{bits} = $1;
              until((($_ = <PNPDATA>) =~ /^# \(IO $io .*/) || eof(PNPDATA)) {
                $profile->{io}[$io]{fixed} = 0;
                if(/^#\s*Minimum IO base address 0x(.*)/) {
                  $profile->{io}[$io]{addr_min} = hex($1);
                }
                elsif(/^#\s*Maximum IO base address 0x(.*)/) {
                  $profile->{io}[$io]{addr_max} = hex($1);
                }
                elsif(/^#\s*IO base alignment (\S*) bytes/) {
                  $profile->{io}[$io]{addr_align} = $1;
                }
                elsif(/^#\s*Number of IO addresses required: (.*)/) {
                  $profile->{io}[$io]{window_size} = $1;
                }
              }
              $io++;
            }
# a fixed IO window definition
            elsif(/^#\s*Fixed IO base address 0x(.*)/) {
              $profile->{io}[$io]{addr_min} = hex($1);
              $profile->{io}[$io]{addr_max} = hex($1);
              $profile->{io}[$io]{addr_align} = 1;
              $profile->{io}[$io]{fixed} = 1;
              until((($_ = <PNPDATA>) =~ /^# \(IO $io .*/) || eof(PNPDATA)) {
                if(/^#\s*Number of IO addresses required: (.*)/) {
                  $profile->{io}[$io]{window_size} = $1;
                }
              }
              $io++;
            }
# a memory range definition (we're NOT handling this yet, I need an example)
            elsif(/^#\s*Memory is/) {
              until((($_ = <PNPDATA>) =~ /^@ \(MEM $mem \(BASE .*/) || eof(PNPDATA)) {
                if(/^#\s*Minimum memory base address 0x(\S*)/) {
                  $profile->{mem}[$mem]{addr_min} = hex($1);
                }
                elsif(/^#\s*Maximum memory base address 0x(\S*)/) {
                  $profile->{mem}[$mem]{addr_max} = hex($1);
                }
                elsif(/^#\s*Range base alignment mask 0xff(\S*) bytes/) {
                  $profile->{mem}[$mem]{addr_align} = hex($1);
                }
                elsif(/^#\s*Range length (\S*) bytes/) {
                  $profile->{mem}[$mem]{window_size} = $1;
                }
              }
              $mem++;
            }
          }
        }
      }
    }
  }
  close(PNPDATA);
  &build_currentrecs($pnp);
  $pnp;
}

sub build_ld_current {
  my($pnp, $brd, $ldn, $prof) = @_;
  my($ld) = $pnp->{board}[$brd]{ld}[$ldn];
  $ld->{current} = {};
  $ld->{current}{profile} = $prof;
  my($k);
  for $k (0 .. $#{$ld->{profile}[$prof]{irq}}) {
    $ld->{current}{irq}[$k] = $ld->{profile}[$prof]{irq}[$k]{irqs}[0];
  }
  for $k (0 .. $#{$ld->{profile}[$prof]{dma}}) {
    $ld->{current}{dma}[$k] = $ld->{profile}[$prof]{dma}[$k]{dmas}[0];
  }
  for $k (0 .. $#{$ld->{profile}[$prof]{io}}) {
    $ld->{current}{io}[$k] = $ld->{profile}[$prof]{io}[$k]{addr_min};
  }
  for $k (0 .. $#{$ld->{profile}[$prof]{mem}}) {
    $ld->{current}{mem}[$k] = $ld->{profile}[$prof]{mem}[$k]{addr_min};
  }
}

sub build_currentrecs {
  my($pnp) = @_;
  if($#{$pnp->{board}} != -1) {
    my($i);
    for $i (1 .. $#{$pnp->{board}}) {
      my($board) = $pnp->{board}[$i];
      my($j);
      for $j ($board->{first_ld} .. $#{$board->{ld}}) {
        build_ld_current($pnp, $i, $j, 0);
      }
    }
  }
}

1;
